#ifndef HARKTOOL4_VALIDATORS_HPP
#define HARKTOOL4_VALIDATORS_HPP

#include <string>
#include <vector>
#include <boost/any.hpp>
#include <boost/program_options.hpp>
#include <boost/assert.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/tokenizer.hpp>

#include "Mode.hpp"
#include "OutputType.hpp"
#include "OutputFormat.hpp"

namespace
{

template <typename T>
class PushBackWithCast
{
    std::vector<T>& vect_;
public:
    PushBackWithCast(std::vector<T>& vect) : vect_(vect) {}
    PushBackWithCast& operator()(const std::string& s)
    {
        namespace po = boost::program_options;
        try
        {
            vect_.push_back(boost::lexical_cast<T>(s));
        }
        catch (const boost::bad_lexical_cast&)
        {
            /* invalid_option_valueのコンストラクタにsを渡すが、
             * sの情報を後で利用する方法はない */
            throw po::invalid_option_value(s);
        }
        return *this;
    }
};

}

namespace harktool4
{

template <typename T> struct CommaSeparatedOption;
/* forward declaration for enum is supported from gcc 4.6
enum class Mode;
enum class OutputType;
enum class OutputFormat;
*/

// custom validators
template <typename T>
void validate(boost::any& v, const std::vector<std::string>& values, CommaSeparatedOption<T>* target, int);
void validate(boost::any& v, const std::vector<std::string>& values, Mode* target, int);
void validate(boost::any& v, const std::vector<std::string>& values, OutputType* target, int);
void validate(boost::any& v, const std::vector<std::string>& values, OutputFormat* target, int);

/** @brief CommaSeparatedOption用バリデータ.
 * カンマ区切りのオプションを分解し、CommaSeparatedOptionを生成する。
 */
template <typename T>
void validate(boost::any& v, const std::vector<std::string>& values, CommaSeparatedOption<T>* target, int)
{
    namespace po = boost::program_options;
    // まだvに割り当てが行われていないことをチェック（エラー時は例外を投げる）
    po::validators::check_first_occurrence(v);
    // vに領域を割り当て
    v = boost::any(CommaSeparatedOption<T>());

    CommaSeparatedOption<T>* p = boost::any_cast<CommaSeparatedOption<T>>(&v);
    BOOST_ASSERT(p);

    const char* const separator_char = ",";
    typedef boost::char_separator<char> separator_t;
    separator_t sep(separator_char);

    // オプションの引数を1つだけ持ってくる
    // （0個だったり、2個以上の場合は例外が投げられる。
    // つまり、このバリデータは単一引数のオプション専用ということ。
    // po::value<CommaSeparatedOption<T>>()->multitoken()の場合には使えない。）
    const std::string& value = po::validators::get_single_string(values);

    PushBackWithCast<T> push_back_with_cast(p->values);

    if (value.find(separator_char) != std::string::npos)
    {
        // separator_char中の文字で*tが表す文字列を区切り、intに変換しつつp->valuesに追加
        boost::tokenizer<separator_t> tok(value, sep);
        std::for_each(tok.begin(), tok.end(), push_back_with_cast);
    }
    else
    {
        // separator_char中の文字が*tに含まれないなら、*tすべてをp->valuesに追加
        push_back_with_cast(value);
    }
}

}

#endif // HARKTOOL4_VALIDATORS_HPP

